﻿using SimpleJSON;
using System;
using System.Linq;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using UnityEngine.EventSystems;
using uFileBrowser;
using Battlehub.RTCommon;
using System.Linq.Expressions;
using UnityThreading;

namespace VAM_Decal_Maker
{
    public class Decal_Maker : MVRScript
    {
        const string pluginName = "DecalMaker";
        const string pluginVersion = "0.6";
        //person script is attatched too
        Atom _parentAtom; //VAM parent
        Atom _shaderAtom; //Obj holding shader
        bool _SetupFinished = false;
        int _queuedBakes = 0;

        public Shader _customDecalShader { get; private set; }
        public Shader _customSpecGlossShader { get; private set; }
        public Shader _customNormShader { get; private set; }
        public Shader _customPackedNormShader { get; private set; }
        public Shader _cutoutShader { get; private set; }
        
        public Dictionary<string, int[]> _TextureIndex { get; private set; }


        public Texture2D _clearTex { get; private set; }
        public DAZCharacterTextureControl _dazCharactertextureControl { get; private set; }
        public DAZMergedSkinV2 _dazSkin { get; private set; }
  

        //UI
        JSONStorableBool _toggleNippleCutout = new JSONStorableBool("Nipple Cutouts ON", true);
        JSONStorableBool _toggleGenitalCutout = new JSONStorableBool("Genital Cutouts ON", true);


        UIDynamicButton _resetButton;
        UIDynamicButton _saveButton;
        UIDynamicButton _clearAllButton;
        UIDynamicButton _unusedButton;
        UIDynamicButton _savePresetButton;
        UIDynamicButton _loadPresetButton;

        UIDynamicToggle _toggleNipple;
        UIDynamicToggle _toggleGenital;

        HeaderPanel _headerPanel;

        public override void Init()
        {
            try
            {
                if (containingAtom.mainController == null)
                {
                    SuperController.LogError("Please add this plugin to a PERSON atom.");
                    return;
                }
                _parentAtom = containingAtom;
                _parentAtom.mainController.interactableInPlayMode = true;
                //load custom asset bundle at get the customshader
                StartCoroutine(CreateAssetAtom("Chokpahi-Asset", "Chokpahi-Asset"));

                _clearAllButton = CreateButton("Clear Everything");
                _unusedButton = CreateButton("Not Used", true);

                _resetButton = CreateButton("Reset Texture");
                _saveButton = CreateButton("Apply Texture", true);

                _loadPresetButton = CreateButton("Load Preset");
                _savePresetButton = CreateButton("Save Preset", true);

                _toggleNipple = CreateToggle(_toggleNippleCutout);
                _toggleGenital = CreateToggle(_toggleGenitalCutout, true);

                RegisterBool(_toggleNippleCutout);
                RegisterBool(_toggleGenitalCutout);
                
                //assign listners
                _clearAllButton.button.onClick.AddListener(ResetAll);

                _unusedButton.button.onClick.AddListener(() =>
                {
                    SuperController.LogError(_dazCharactertextureControl.uvSetName);
                });

                _saveButton.button.onClick.AddListener(() =>
                {
                    TempManager.DestoryAll();
                    _headerPanel.Bake();
                });

                _savePresetButton.button.onClick.AddListener(() =>
                {
                    string presets = "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Presets/";
                    if (!MVR.FileManagementSecure.FileManagerSecure.DirectoryExists(presets))
                    {
                        //HACK
                        _dazSkin.dazMesh.CreateFolderIfNeeded(presets);
                    }
                    SuperController.singleton.fileBrowserUI.defaultPath = presets;
                    SuperController.singleton.fileBrowserUI.showFiles = true;
                    SuperController.singleton.fileBrowserUI.showDirs = true;
                    SuperController.singleton.fileBrowserUI.SetTextEntry(true);
                    SuperController.singleton.fileBrowserUI.Show(SavePresetCallback);

                });

                _loadPresetButton.button.onClick.AddListener(() =>
                {
                    string presets = "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Presets";

                    SuperController.singleton.fileBrowserUI.defaultPath = presets;
                    SuperController.singleton.fileBrowserUI.SetTitle("Select Load Decal Preset");
                    SuperController.singleton.fileBrowserUI.showFiles = true;
                    SuperController.singleton.fileBrowserUI.showDirs = true;
                    SuperController.singleton.fileBrowserUI.SetTextEntry(false);
                    SuperController.singleton.fileBrowserUI.Show(PresetLoadCallBack);
                });
               
                byte[] tmppng = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/Clear.png");
                _clearTex = new Texture2D(1, 1, TextureFormat.RGBA32, false);
                _clearTex.LoadImage(tmppng);

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        void DebugSave(Texture2D tempTex, string path)
        {
            //byte[] bytes = tempTex.EncodeToPNG();
           // MVR.FileManagementSecure.FileManagerSecure.WriteAllBytes(path, bytes);
        }
        //MacGruber Discord 03/26/2020
        //Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
        public static string GetPluginPath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }

        // Get path prefix of the package that contains our plugin.
        public static string GetPackagePath(MVRScript self)
        {
            string filename = GetPluginPath(self);
            int idx = filename.IndexOf(":/");
            if (idx >= 0)
                return filename.Substring(0, idx + 2);
            else
                return string.Empty;
        }
        //"project-canyon12/05/2018 DISCORD"
        //"I'm going on how I understand Async/Await works in normal C#, but I think the principle is the same"
        //tweaked for my use case
        delegate void AtomCreatedEvent(Atom atom);
        IEnumerator CreateAssetAtom(string uniqueAtomName, string nameOfGameObjectToFind)
        {
            yield return new WaitForSeconds(0.5f);

            Atom createdAtom = GetAtomById(uniqueAtomName);
            if (createdAtom == null)
            {
                yield return SuperController.singleton.AddAtomByType("CustomUnityAsset", uniqueAtomName, true);

                createdAtom = GetAtomById(uniqueAtomName);
            }

            JSONStorable storable = createdAtom.GetStorableByID("asset");
            storable.SetUrlParamValue("assetUrl", GetPackagePath(this) + @"Custom\Assets\Chokaphi\chokpahi-decal.assetbundle");
            storable.SetStringChooserParamValue("assetName", "Assets/Chokpahi-Decal.unity");

            yield return new WaitWhile(() => GameObject.Find(nameOfGameObjectToFind) == null);

            SuperController.LogError("Asset loaded");
            _shaderAtom = createdAtom;

            yield return new WaitForSeconds(0.5f);
            Shader[] renderers = Resources.FindObjectsOfTypeAll<Shader>();
            foreach (Shader shader in renderers)
            {
                if (shader.name == "Custom/Unlit-Decal2")
                {
                    SuperController.LogError("ShaderFind " + Shader.Find(shader.name) + " Custom Shader is " + shader.name);
                    _customDecalShader = shader;
                }
                if (shader.name == "Custom/Unlit-Cutout2")
                {
                    SuperController.LogError("ShaderFind " + Shader.Find(shader.name) + " Cutout Shader is " + shader.name);
                    _cutoutShader = shader;
                }
                if (shader.name == "Custom/Unlit-Specular")
                {
                    SuperController.LogError("ShaderFind " + Shader.Find(shader.name) + " Spec-Gloss Shader is " + shader.name);
                    _customSpecGlossShader = shader;
                }
                if (shader.name == "Custom/Unlit-NormalMap")
                {
                    SuperController.LogError("ShaderFind " + Shader.Find(shader.name) + " Normal Shader is " + shader.name);
                    _customNormShader = shader;
                }
                if (shader.name == "Custom/Unlit-PackedNormalMap")
                {
                    SuperController.LogError("ShaderFind " + Shader.Find(shader.name) + " PackedNorm Shader is " + shader.name);
                    _customPackedNormShader = shader;
                }
            }

        }

        // Update is called with each rendered frame by Unity
        void Update()
        {
            try
            {
                //need to delay creation on Sceen reloads till VAM has finished loading
                if (_parentAtom && _customDecalShader && _dazCharactertextureControl == null)
                {
                    _dazCharactertextureControl = _parentAtom.GetComponentInChildren<DAZCharacterTextureControl>();
                    if (_dazCharactertextureControl != null)
                    {
                        _dazSkin = _dazCharactertextureControl.GetComponent<DAZMergedSkinV2>();
                        if (_dazSkin != null)
                        {
                            if (_SetupFinished)
                            {
                                _headerPanel.UpdateSelectionPanels();
                                StartCoroutine(ApplyDecalsThreaded(1f));
                            }
                            else
                            {
                                Setup();
                            }
                        }
                    }
                }

            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        //this is called on SKIN changes from the Update Event
        public void Setup()
        {
            try
            {
                _TextureIndex = new Dictionary<string, int[]>()
                {
                    {"torso" , _dazCharactertextureControl.torsoMaterialNums },
                    { "face" , _dazCharactertextureControl.faceMaterialNums },
                    { "limbs", _dazCharactertextureControl.limbMaterialNums },
                    { "genitals", _dazCharactertextureControl.genitalMaterialNums },
                };

                //create Header Panel
                _headerPanel = new HeaderPanel(this);

                _resetButton.button.onClick.AddListener(_headerPanel.ResetTextures);


                string path = "Custom/Scripts/Chokaphi/OUT/";

                _headerPanel.UpdateSelectionPanels();
    
                UIDynamicButton NewPanelButton = CreateButton(" LAST LEFT ELEMENT");
                UIDynamicButton UnusedButton2 = CreateButton("LAST RIGHT ELEMENT", true);

                //this is hacky but sets the UI to look nice on start.
                _headerPanel.SelectionPanelSpec.TorsoPanel.Button.onClick.Invoke();
                _headerPanel.SelectionPanelGloss.TorsoPanel.Button.onClick.Invoke();
                _headerPanel.SelectionPanelNorm.TorsoPanel.Button.onClick.Invoke();


                _headerPanel.decalButton.button.onClick.Invoke();
                _headerPanel.SelectionPanelDecal.TorsoPanel.Button.onClick.Invoke();


                _SetupFinished = true;
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }
        }

        public void UpdateSelectionPanels(SelectionPanel selectionPanel)
        {
            foreach (var panel in selectionPanel.Selections)
            {
                int index = _TextureIndex[panel.Key].FirstOrDefault();
                string slot = panel.Value.MaterialSlot;

                //test if we have Source Texture
                if (panel.Value.MaterialSlot != "_DecalTex" && (Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot) == null)
                {
                    //SuperController.LogError("No source texture" + slot + " " + panel.Key);
                    return;
                }

                Texture2D temp;// = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot);

                if (panel.Value.MaterialSlot == "_DecalTex")
                {
                    temp = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture("_MainTex");
                }
                else if (panel.Value.MaterialSlot == "_BumpMap" && CheckIfPackedNormal((Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot)))
                {
                    temp = GpuCombine(_dazSkin.GPUmaterials[index].GetTexture(slot), _customPackedNormShader, panel.Value.MaterialSlot, Color.white, null, true);
                }
                else
                {
                    temp = (Texture2D)_dazSkin.GPUmaterials[index].GetTexture(slot);
                }

                panel.Value.image.material.SetTexture("_MainTex" , temp);
                panel.Value.Original = _dazSkin.GPUmaterials[index].GetTexture(panel.Value.MaterialSlot);
                //SuperController.LogError(panel.Key + " " + panel.Value.MaterialSlot + " " + slot + " " + index);
            }

        }

        private void ResetAll()
        {
            _headerPanel.ResetTextures();

            foreach (var v in _headerPanel.selectionPanels)
            {
                foreach (var selPanel in    v.Value.Selections)
                {
                    int count = selPanel.Value.DecalPanels.Count;
                    for (int i = 0; i < count; i++)
                    {
                        selPanel.Value.RemoveDecalPanel();
                    }
                }
            }
        }

        public void ResetTextures(ImagePanel panel)
        {
            foreach (int num in _TextureIndex[panel.TextureSlot])
            {
                _dazSkin.GPUmaterials[num].SetTexture(panel.MaterialSlot, panel.Original);
            }
            TempManager.DestoryAll();
        }
        
        //used when restoring from json or profile.
        //make sure all images are loaded before applying
        private IEnumerator ApplyDecalsThreaded(float delay = 0.5f)
        {
            yield return new WaitForSeconds(delay);

            foreach (var panels in _headerPanel.selectionPanels)
            {
                foreach (var selPanel in panels.Value.Selections)
                {
                    if (selPanel.Value.DecalPanels.Count > 0)
                    {
                        foreach (DecalPanel d in selPanel.Value.DecalPanels)
                        {
                            yield return new WaitWhile(() => d.ImagePanel.ImageLoading == true);
                        }
                    }
                }
            }
            TempManager.DestoryAll();
            _headerPanel.Bake();
        }

        public bool CheckIfPackedNormal(Texture2D tempTex)
        {
            if (tempTex == null)
            {
                return false;
            }

            try 
            {
                tempTex.GetPixels();
            }
            catch 
            {
                tempTex = GetUnreadableTexture(tempTex, true);
            }

            for (int x = 0; x < tempTex.width; x++)
            {
                for (int y = 0; y < tempTex.height; y++)
                {
                    if (tempTex.GetPixel(x, y).b != 0)
                    {
                        return false;
                    }
                }
            }
            return true;
        }
        
        public void Bake(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
           
            if (!_customDecalShader || !_cutoutShader)
            {
                SuperController.LogError("CUSTOM SHADER NOT FOUND");
                return;
            }
            _queuedBakes++;
            StartCoroutine(BakeThreaded(DecalPanels, TextureSlot, MaterialSlot));
        }
        private IEnumerator BakeThreaded(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            yield return new WaitForEndOfFrame();
            switch (MaterialSlot)
            {
                case "_DecalTex":
                    StartCoroutine(BakeDecal(DecalPanels, TextureSlot, MaterialSlot));
                    break;
                case "_SpecTex":
                case "_GlossTex":
                    StartCoroutine(BakeSpecular(DecalPanels, TextureSlot, MaterialSlot));
                    break;
                case "_BumpMap":
                    StartCoroutine(BakeNormal(DecalPanels, TextureSlot, MaterialSlot));
                    break;
            }

            yield return new WaitWhile(() => _queuedBakes > 0);
            TempManager.DestoryTempObjs();
        }

        private IEnumerator BakeDecal(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            string path = "Custom/Scripts/Chokaphi/OUT/";
            int count = 0;
         
            Texture2D tempTex = TempManager.CopyTexture(_clearTex);
            foreach (DecalPanel d in DecalPanels)
            {
                if (d.ImagePanel.image.mainTexture != null)
                {
                    yield return new WaitForEndOfFrame();
                    tempTex = GpuCombine(tempTex, _customDecalShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture);
                    DebugSave(tempTex, path + TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                    count++;
                }
            }
            //apply cutouts
            SuperController.LogError(TextureSlot + MaterialSlot);
            if (TextureSlot == "torso" && _toggleNippleCutout.val && _cutoutShader != null)
            {
                yield return new WaitForEndOfFrame();
                byte[] png = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/" + _dazCharactertextureControl.uvSetName + ".png");
                Texture2D alphaTex = TempManager.TempTexture2D(1, 1, TextureFormat.RGBA32, false);
                alphaTex.LoadImage(png);
                tempTex = GpuClip(tempTex, alphaTex);
            }
            if (TextureSlot == "genitals" && _toggleGenitalCutout.val && _cutoutShader != null)
            {
                yield return new WaitForEndOfFrame();
                byte[] png = MVR.FileManagementSecure.FileManagerSecure.ReadAllBytes(GetPackagePath(this) + "Custom/Scripts/Chokaphi/VAM_Decal_Maker/Cutout/_FemaleGenitals.png");
                Texture2D alphaTex = TempManager.TempTexture2D(1, 1, TextureFormat.RGBA32, false);
                alphaTex.LoadImage(png);
                tempTex = GpuClip(tempTex, alphaTex);
            }

            foreach (int num in _TextureIndex[TextureSlot])
            {
                yield return new WaitForEndOfFrame();
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }
        //handles Specular and Gloss linear textures
        private IEnumerator BakeSpecular(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            //Texture2D tempTex = (Texture2D)_dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].GetTexture(MaterialSlot);
            Texture2D tempTex = null;
            string path = "Custom/Scripts/Chokaphi/OUT/";
            int count = 0;
            foreach (DecalPanel d in DecalPanels)
            {
                if (d.ImagePanel.image.mainTexture != null)
                {
                    if (tempTex == null)
                    {
                        tempTex = TempManager.CopyTexture((Texture2D)d.ImagePanel.image.mainTexture, true);
                    }
                    else
                    {
                        //yield return new WaitForEndOfFrame();
                        tempTex = GpuCombine(tempTex, _customSpecGlossShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture, true);
                    }
                    DebugSave(tempTex, path + TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                    count++;
                }
            }
            foreach (int num in _TextureIndex[TextureSlot])
            {
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }
       
        private IEnumerator BakeNormal(List<DecalPanel> DecalPanels, string TextureSlot, string MaterialSlot)
        {
            if (!_dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].HasProperty("_BumpMap"))
            {
                SuperController.LogError("NO bump map property " + TextureSlot + " " + _dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].shaderKeywords);
                yield break; 
            }

            SuperController.LogError("queue amount = " + _queuedBakes);
            string path = "Custom/Scripts/Chokaphi/OUT/";
            Texture2D tempTex = (Texture2D)_dazSkin.GPUmaterials[_TextureIndex[TextureSlot][0]].GetTexture(MaterialSlot);
         
            if (tempTex != null && CheckIfPackedNormal(tempTex))
            {
                tempTex = GpuCombine(tempTex, _customPackedNormShader, "_BumpTex", Color.white);
            }
           
            foreach (DecalPanel d in DecalPanels)
            {
                int count = 0;
                if (d.ImagePanel.image.mainTexture != null)
                {
                    if (tempTex == null)
                    {
                        tempTex = TempManager.CopyTexture((Texture2D)d.ImagePanel.image.mainTexture, true);
                    }
                    else
                    { 
                        //yield return new WaitForEndOfFrame();
                        tempTex = GpuCombine(tempTex, _customNormShader, MaterialSlot, d.Color, d.ImagePanel.image.mainTexture, true);
                    }
                    DebugSave(tempTex, path + TextureSlot + "-" + MaterialSlot + " " + count + ".png");
                }
            }
           
            foreach (int num in _TextureIndex[TextureSlot])
            {
                _dazSkin.GPUmaterials[num].SetTexture(MaterialSlot, tempTex);
            }

            TempManager.MakeSessionObj(tempTex);
            _queuedBakes--;
            yield break;
        }

        public void PresetLoadCallBack(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;

            JSONClass jc = this.LoadJSON(path) as JSONClass;
            PerformLoad(jc);
        }

        public void SavePresetCallback(string path)
        {
            if (string.IsNullOrEmpty(path))
                return;

            JSONClass json = GetJSON();
            this.SaveJSON(json, path + "." + DSGN + ".DecalMakerPreset.json");
            SuperController.singleton.DoSaveScreenshot(path + "."+ DSGN +".DecalMakerPreset.json");
        }

        //SAVE
        private string DSGN = "dsgn";
        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance);
            jc["SaveVersion"] = "1";

            //save
            jc["Decal"] = _headerPanel.SelectionPanelDecal.Save();
            jc["Specular"] = _headerPanel.SelectionPanelSpec.Save();
            jc["Gloss"] = _headerPanel.SelectionPanelGloss.Save();
            jc["Normal"] = _headerPanel.SelectionPanelNorm.Save();


            jc["Nipple Cutouts ON"] = _toggleNippleCutout.val.ToString();
            jc["Genital Cutouts ON"] = _toggleGenitalCutout.val.ToString();

            return jc;
        }

        //LOAD
        public override void RestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, JSONArray presetAtoms = null, bool setMissingToDefault = true)
        {
            base.RestoreFromJSON(jc, restorePhysical, restoreAppearance, presetAtoms);
            StartCoroutine(PerformLoadThreaded(jc));
        }
        //called once we have everything loaded. RestorefromJSON is too early
        private IEnumerator PerformLoadThreaded(JSONClass jc)
        {
            yield return new WaitWhile(() => _SetupFinished == false && _customDecalShader == null);
            PerformLoad(jc);
        }

        private void PerformLoad(JSONClass jc)
        {
            if (jc == null)
                return;
            //convert to new file version if save is old
            if (jc["SaveVersion"] == null)
                jc = ConvertSaveToV1(jc);


            if (jc["Nipple Cutouts ON"] != null)
            {
                _toggleNippleCutout.SetVal(bool.Parse(jc["Nipple Cutouts ON"]));
            }

            if (jc["Genital Cutouts ON"] != null)
            {
                _toggleGenitalCutout.SetVal(bool.Parse(jc["Genital Cutouts ON"]));
            }

            foreach (var panel in _headerPanel.selectionPanels)
            {
                if (jc[panel.Key] != null)
                {
                    foreach (var selPanel in panel.Value.Selections)
                    {
                        foreach (JSONClass saveJSON in jc[panel.Key][selPanel.Key] as JSONArray)
                        {
                            DecalPanel d =   panel.Value.AddDecalPanel(selPanel.Key);
                            if (saveJSON["Alpha"] != null)
                            {
                                d.sliderDynamic.slider.value = (float.Parse(saveJSON["Alpha"]));
                            }
                            if (saveJSON["H"] != null)
                            {
                                float h = float.Parse(saveJSON["H"]);
                                float s = float.Parse(saveJSON["S"]);
                                float v = float.Parse(saveJSON["V"]);
                                d.jcolor.SetVal(h, s, v);
                            }
                            if (saveJSON["Path"] != null)
                            {
                                d.BrowserCallBack(saveJSON["Path"]);
                            }
                        }
                    }
                }
            }
           
            StartCoroutine(ApplyDecalsThreaded());
        }
        //upgrade old saves to new format
        private JSONClass ConvertSaveToV1(JSONClass jc)
        {
            List<string> keys = new List<string>() { "torso", "genitals","face","limbs" };
            JSONClass dc = new JSONClass();
            foreach (string key in keys)
            {
                if (jc[key] != null)
                {
                    dc.Add(key, jc[key]);
                }
                jc.Remove(key);
            }
            jc["Decal"] = dc;
        
            return jc;
        }

        //VAM uses System Drawing see ImageLoaderThreaded Process()
        public Texture2D GetTextureFromSurfaceShader(Material mat, int width, int height)
        {
            //Create render texture:
            RenderTexture temp = new RenderTexture(width, height, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
            temp.DiscardContents();
            //Create a Quad:
            GameObject quad = GameObject.CreatePrimitive(PrimitiveType.Quad);
            MeshRenderer rend = quad.GetComponent<MeshRenderer>();

            rend.material = mat;

            // Vector3 quadScale = quad.transform.localScale / (float)((Screen.height / 2.0) / Camera.main.orthographicSize);
            Vector3 quadScale = quad.transform.localScale;
            quad.transform.position = Vector3.forward;

            //Setup camera:
            GameObject camGO = new GameObject("CaptureCam");
            Camera cam = camGO.AddComponent<Camera>();
            cam.renderingPath = RenderingPath.Forward;
            cam.orthographic = true;
            cam.clearFlags = CameraClearFlags.SolidColor;
            cam.backgroundColor = new Color(1, 1, 1, 0);
            if (cam.rect.width < 1 || cam.rect.height < 1)
            {
                cam.rect = new Rect(cam.rect.x, cam.rect.y, 1, 1);
            }
            cam.orthographicSize = 0.5f;
            cam.rect = new Rect(0, 0, quadScale.x, quadScale.y);
            cam.aspect = quadScale.x / quadScale.y;
            cam.targetTexture = temp;
            cam.allowHDR = false;


            //Capture image and write to the render texture:
            cam.Render();
            temp = cam.targetTexture;

            //Apply changes:
            Texture2D newTex = new Texture2D(temp.width, temp.height, TextureFormat.RGBA32, true, false);
            RenderTexture.active = temp;
            newTex.ReadPixels(new Rect(0, 0, temp.width, temp.height), 0, 0);
            newTex.Apply();

            //Clean up:
            RenderTexture.active = null;
            temp.Release();
            Destroy(quad);
            Destroy(camGO);

            return newTex;
        }

        //Uses shader to clip part of one texture and applys it to another based on the alpha of a third texture.
        //Used to Apply "Clean" nipples and genital areas after decals have been applied.
        //This Alpha control will be Character UV specific. So Victoria has a diffrent nipple area than Olympia etc..
        public Texture2D GpuClip(Texture mainTex, Texture alphaTex)
        {
            //SuperController.LogError("GPU Clip textures " + _bitsShader.name);

            int h = mainTex.width;
            int w = mainTex.height;

            RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);

            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(_cutoutShader);
            material.mainTexture = mainTex;
            material.SetTexture("_Alpha", alphaTex);

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        public Texture2D GpuCombine(Texture mainTex, Shader shader, string materialSlot, Color color, Texture overlayTex = null, bool linear = false)
        {
            int h = mainTex.width;
            int w = mainTex.height;

            RenderTextureReadWrite renderTexConversion = RenderTextureReadWrite.sRGB;
            if (linear)
            {
                renderTexConversion = RenderTextureReadWrite.Linear;
            }

             RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, renderTexConversion);
            //tmp.antiAliasing = 8;
            //Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(shader);
            material.mainTexture = mainTex;
                
            if (materialSlot == "_DecalTex")
            {
                material.SetVector("_DecalColor", color); 
            }
            if (overlayTex)
            {
                material.SetTexture(materialSlot, overlayTex);
            }

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false, linear);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0, false);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        //old method
        public Texture2D GpuCombine(Texture mainTex, Texture decalTex, Color color)
        {
            int h = mainTex.width;
            int w = mainTex.height;

            RenderTexture tmp = RenderTexture.GetTemporary(w, h, 0, RenderTextureFormat.ARGB32, RenderTextureReadWrite.sRGB);
            //tmp.antiAliasing = 8;
            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;

            Material material = TempManager.TempMaterial(_customDecalShader);
            material.mainTexture = mainTex;
            material.SetTexture("_DecalTex", decalTex);
            material.SetVector("_DecalColor", color);

            Graphics.Blit(mainTex, tmp, material);
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(w, h, TextureFormat.RGBA32, false, false);
            newTexture.ReadPixels(new Rect(0, 0, w, h), 0, 0, false);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            RenderTexture.ReleaseTemporary(tmp);

            return newTexture;
        }
        //From https://support.unity3d.com/hc/en-us/articles/206486626-How-can-I-get-pixels-from-unreadable-textures-
        public Texture2D GetUnreadableTexture(Texture2D texture, bool linear = false)
        {
            RenderTextureReadWrite renderTexConversion = RenderTextureReadWrite.sRGB;
            if (linear)
            {
                renderTexConversion = RenderTextureReadWrite.Linear;
            }
            //SuperController.LogError("Get Unreadable texture");
            RenderTexture tmp = RenderTexture.GetTemporary(texture.width, texture.height, 0, RenderTextureFormat.ARGB32, renderTexConversion);

            // Blit the pixels on texture to the RenderTexture
            Graphics.Blit(texture, tmp);
            // Backup the currently set RenderTexture
            RenderTexture previous = RenderTexture.active;
            // Set the current RenderTexture to the temporary one we created
            RenderTexture.active = tmp;

            Texture2D newTexture = TempManager.TempTexture2D(texture.width, texture.height, TextureFormat.RGBA32, false, linear);

            newTexture.ReadPixels(new Rect(0, 0, tmp.width, tmp.height), 0, 0);
            newTexture.Apply();

            // Reset the active RenderTexture
            RenderTexture.active = previous;
            // Release the temporary RenderTexture
            RenderTexture.ReleaseTemporary(tmp);
            //SuperController.LogError("Finish Getting Unreadable texture");
            return newTexture;
        }

    }

    public class DecalPanel
    {
        public GameObject gameObject = new GameObject("DecalPanel");
        public ImagePanel ImagePanel { get; private set; }

        public Color Color { get; private set; } = Color.white;

        public JSONStorableColor jcolor { get; private set; }

        public UIDynamicSlider sliderDynamic { get; private set; }

        public Action<string> LastDir;
        public DecalPanel(Transform anchor, float width, float height, Color? color, MVRScript mVR)
        {
            gameObject.transform.SetParent(anchor, false);
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = color ?? new Color(0, 0, 0, 0f);
            panelImage.rectTransform.sizeDelta = new Vector2(width, height);

            //Image panel
            GameObject imageObject = new GameObject("ImagePanel");
            imageObject.transform.SetParent(gameObject.transform, false);

            ImagePanel = new ImagePanel();
            ImagePanel.gameObject.transform.SetParent(gameObject.transform, false);
            ImagePanel.gameObject.transform.localPosition = new Vector3(panelImage.rectTransform.rect.xMax - ImagePanel.image.rectTransform.rect.xMax - 10, 0, 0);


            Transform sliderPrefab = GameObject.Instantiate<Transform>(mVR.manager.configurableSliderPrefab);
            sliderPrefab.SetParent(gameObject.transform, false);
            sliderPrefab.localPosition = Vector3.zero;

            sliderDynamic = sliderPrefab.GetComponent<UIDynamicSlider>();
            sliderDynamic.transform.localPosition = new Vector2(0, -80);
            sliderDynamic.quickButtonsEnabled = false;
            sliderDynamic.rangeAdjustEnabled = false;
            sliderDynamic.defaultButtonEnabled = false;
            sliderDynamic.label = "";

            Image sliderBackground = sliderDynamic.GetComponentsInChildren<Image>().Where(x => x.name == "Panel").First();
            Color c = sliderBackground.color;
            c.a = 0.1f;
            sliderBackground.color = c;

            sliderDynamic.slider.onValueChanged.AddListener(SetAlpha);



            sliderDynamic.slider.minValue = 0;
            sliderDynamic.slider.maxValue = 1;
            sliderDynamic.slider.value = 1;


            GameObject colorObject = new GameObject("ColorPanel");
            colorObject.transform.SetParent(gameObject.transform, false);
            colorObject.transform.localScale = new Vector3(.55f, .55f, .55f);
            colorObject.transform.localPosition = new Vector2(-350, 50);


            Image colorImage = colorObject.AddComponent<Image>();
            colorImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            colorImage.color = new Color(0, 1, 1, 0f);
            colorImage.rectTransform.sizeDelta = new Vector2(600, 300);
            colorImage.transform.localPosition = new Vector2(-350, 50);

            Transform colorPickerPrefab = GameObject.Instantiate<Transform>(mVR.manager.configurableColorPickerPrefab);
            colorPickerPrefab.SetParent(colorImage.transform, false);
            colorPickerPrefab.localPosition = Vector3.zero;

            UIDynamicColorPicker colorPickerDynamic = colorPickerPrefab.GetComponent<UIDynamicColorPicker>();
            colorPickerDynamic.showLabel = false;
            HSVColor hsvc = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
            jcolor = new JSONStorableColor("FooColor", hsvc, ColorCallback);
            jcolor.colorPicker = colorPickerDynamic.colorPicker;

        }

        private void SetAlpha(float value)
        {
            ImagePanel.SetAlpha(value);
            Color temp = Color;
            temp.a = value;
            Color = temp;
            jcolor.SetColor(temp);
        }

        private void ColorCallback(JSONStorableColor jcolor)
        {
            ImagePanel.SetColor(jcolor.colorPicker.currentColor);
            Color = jcolor.colorPicker.currentColor;
        }
        public void BrowserCallBack(string path)
        {
            //SuperController.LogError("Path is " + path);
            if (string.IsNullOrEmpty(path))
            {
                //LastDir = null;
            }
            else
            {
                LastDir(MVR.FileManagementSecure.FileManagerSecure.GetDirectoryName(path));
                ImagePanel.LoadPNG(path, ImagePanel);
            }
        }

        public JSONClass Save()
        {
            JSONClass decalSaveJSON = new JSONClass();
            decalSaveJSON["H"] = jcolor.val.H.ToString();
            decalSaveJSON["S"] = jcolor.val.S.ToString();
            decalSaveJSON["V"] = jcolor.val.V.ToString();
            decalSaveJSON["Alpha"] = ImagePanel.image.color.a.ToString();

            if (!string.IsNullOrEmpty(ImagePanel.Path))
            {
                string relativePath = MVR.FileManagementSecure.FileManagerSecure.NormalizePath(ImagePanel.Path);
                decalSaveJSON["Path"] = relativePath;
            }
            return decalSaveJSON;
        }
    }          


    public class HeaderPanel
    {
        public GameObject gameObject = new GameObject("Header");

        public SelectionPanel ActivePanel { get; set; }

        public SelectionPanel SelectionPanelDecal { get; private set; }
        public SelectionPanel SelectionPanelSpec { get; private set; }
        public SelectionPanel SelectionPanelGloss { get; private set; }
        public SelectionPanel SelectionPanelNorm { get; private set; }
        public MyButton decalButton { get; private set; }
        public MyButton specButton { get; private set; }
        public MyButton glossButton { get; private set; }
        public MyButton normButton { get; private set; }

        private List<MyButton> buttons { get; set; }

        public Dictionary<string, SelectionPanel> selectionPanels { get; private set; }

        private Decal_Maker _DM;
        public HeaderPanel(Decal_Maker DM)
        {
            _DM = DM;
            //create a UI Spacer to Hold our Panel
            UIDynamic spacerLeft = DM.CreateSpacer(false);
            spacerLeft.height = 50;
            UIDynamic spacerRight = DM.CreateSpacer(true);
            spacerRight.height = 50;

            //Image i = spacerLeft.gameObject.AddComponent<Image>();
            //i.material = new Material(Shader.Find("UI/Default-Overlay"));
            //i.material.color = Color.green;

            gameObject.transform.SetParent(spacerLeft.transform, false);
            //center panel in UI
            gameObject.transform.localPosition += new Vector3(270, 10, 0);
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = new Color(.5f, .5f, .5f, 1);
            panelImage.rectTransform.sizeDelta = new Vector2(1050, 70);

            GridLayoutGroup glg = gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2(200, 50);

            glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(10, 10, 10, 10);
            glg.spacing = new Vector2(10, 0);


            decalButton = new MyButton("Decal", new Color(1, 0.3f, 0.4f));
            decalButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            specButton = new MyButton("Specular", new Color(0.95f, 0.25f, 0.91f));
            specButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            glossButton = new MyButton("Gloss", new Color(0.2f, .98f, 0.2f));
            glossButton.gameObject.transform.SetParent(gameObject.transform, false);
            
            normButton = new MyButton("Normal", new Color(0.2f, 0.9f, .9f));
            normButton.gameObject.transform.SetParent(gameObject.transform, false);

            buttons = new List<MyButton>() { decalButton, specButton, glossButton, normButton };

            SelectionPanelDecal = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_DecalTex",
                isNormalMap = false,
                linear = false
            };


            SelectionPanelSpec = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_SpecTex",
                isNormalMap = false,
                linear = true
            };

            SelectionPanelGloss = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_GlossTex",
                isNormalMap = false,
                linear = true
            };

            SelectionPanelNorm = new SelectionPanel(1050, 300, new Color(1, 1, 1, 1f), DM)
            {
                MaterialSlot = "_BumpMap",
                isNormalMap = true,
                linear = true

            };
            //now finalize panels
            SelectionPanelDecal.CreatePanels();
            SelectionPanelSpec.CreatePanels();
            SelectionPanelGloss.CreatePanels();
            SelectionPanelNorm.CreatePanels();


            SelectionPanelDecal.PanelName = "Decal";
            SelectionPanelSpec.PanelName = "Specular";
            SelectionPanelGloss.PanelName = "Gloss";        
            SelectionPanelNorm.PanelName = "Normal";

            SelectionPanelDecal.ActivateWindow(true);
            SelectionPanelSpec.ActivateWindow(false);
            SelectionPanelGloss.ActivateWindow(false);
            SelectionPanelNorm.ActivateWindow(false);

            ActivePanel = SelectionPanelDecal;

            selectionPanels = new Dictionary<string, SelectionPanel> { { "Decal", SelectionPanelDecal }, { "Specular", SelectionPanelSpec }, { "Gloss", SelectionPanelGloss }, { "Normal", SelectionPanelNorm } };
            foreach (MyButton uB in buttons)
            {
                uB.button.onClick.AddListener(() =>
                {
                    foreach (MyButton b in buttons)
                    {
                        if (b.buttonText == uB.buttonText)
                        {
                            b.Selected(true);
                            ActivePanel = selectionPanels[b.buttonText.text];
                            ActivePanel.ActivateWindow(true);
                        }
                        else
                        {
                            b.Selected(false);
                            ActivePanel = selectionPanels[b.buttonText.text];
                            ActivePanel.ActivateWindow(false);
                        }
                    }
                });
            }
        }

        public void UpdateSelectionPanels()
        {
            selectionPanels["Decal"].UpdateSelectionPanels();
            selectionPanels["Specular"].UpdateSelectionPanels();
            selectionPanels["Gloss"].UpdateSelectionPanels();
            selectionPanels["Normal"].UpdateSelectionPanels();
        }

        public void ResetTextures()
        {
            SelectionPanelDecal.ResetTextures();
            SelectionPanelSpec.ResetTextures();
            SelectionPanelGloss.ResetTextures();
            SelectionPanelNorm.ResetTextures();
        }

        public void Bake()
        {
            SelectionPanelDecal.Bake();
            SelectionPanelSpec.Bake();
            SelectionPanelGloss.Bake();
            SelectionPanelNorm.Bake();
        }
 
    }

    public class MyButton
    {
        public GameObject gameObject = new GameObject("MyButton");
        public Image image { get; set; }

        public Button button { get; set; }
        public Text buttonText { get; set; }

        private DefaultControls.Resources resources = new DefaultControls.Resources();

        private Color normalColor;
        private Color unselectedColor;

        public void Selected(bool value)
        {
            if (value)
            {
                ColorBlock colors = button.colors;
                colors.normalColor = normalColor;
                button.colors = colors;
            }
            else
            {
                ColorBlock colors = button.colors;
                colors.normalColor = unselectedColor;
                button.colors = colors;
            }
        }

        public MyButton(string label = "Button", Color? color = null)
        {
            
            //transform = gameObject.transform;
            RectTransform rt = gameObject.AddComponent<RectTransform>();
            rt.sizeDelta = new Vector2(160, 30);

            GameObject textGO = new GameObject("Text");
            textGO.transform.SetParent(gameObject.transform, false);

            RectTransform tRT = textGO.AddComponent<RectTransform>();
            tRT.anchorMin = Vector2.zero;
            tRT.anchorMax = Vector2.one;
            tRT.sizeDelta = Vector2.zero;

            buttonText = textGO.AddComponent<Text>();
            buttonText.text = label;
            buttonText.alignment = TextAnchor.MiddleCenter;
            buttonText.color = Color.black;
           
            buttonText.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
            buttonText.fontSize = 30;

            image = gameObject.AddComponent<Image>();
            //image.color = color ?? Color.green;

            normalColor = color ?? new Color(0.6f,0,0,1);
            button = gameObject.AddComponent<Button>();
           
            //convert color to HSV to allow easier lighten/darken operation
            float H, S, V;
            Color.RGBToHSV(normalColor, out H, out S, out V);
            ColorBlock colors = button.colors;
            colors.highlightedColor = Color.HSVToRGB(H, S, V * 1.4f);
            colors.normalColor = normalColor;
            colors.pressedColor = normalColor;// Color.HSVToRGB(H, S, V * 1.2f);
            unselectedColor = Color.HSVToRGB(H, S, V * 0.5f);

            button.colors = colors;
        }
    }

    public class SelectionPanel
    {
        public GameObject gameObject = new GameObject("Panel");

        public string PanelName { get; set; }
        public UIDynamicButton AddButton { get; private set; }
        public UIDynamicButton RemoveButton { get; private set; }

        public Dictionary<string, ImagePanel> Selections { get; private set; }
        public ImagePanel ActivePanel { get; private set; }
        public ImagePanel TorsoPanel { get; private set; }
        public ImagePanel GenitalPanel { get; private set; }
        public ImagePanel FacelPanel { get; private set; }  
        public ImagePanel LimbsPanel { get; private set; }

        public string MaterialSlot { get; set; }
        public bool isNormalMap { get; set; } = false;
        public bool linear { get; set; } = false;


        private Decal_Maker _mVR;
        private string _lastDir;
        private UIDynamic spacerLeft;
        private UIDynamic spacerRight;
        
        public SelectionPanel(float width, float height, Color? color, Decal_Maker mVR)
        {
            //create a UI Spacer to Hold our Panel
            spacerLeft = mVR.CreateSpacer(false);
            spacerLeft.height = 300;
            spacerRight = mVR.CreateSpacer(true);
            spacerRight.height = 300;

            //Image i = spacerLeft.gameObject.AddComponent<Image>();
            //i.material = new Material(Shader.Find("UI/Default-Overlay"));
            //i.material.color = Color.blue;

            _mVR = mVR;

             //create buttons to add/remove panels
            AddButton = mVR.CreateButton("Add Image");
            RemoveButton = mVR.CreateButton("Remove Image", true);


            gameObject.transform.SetParent(spacerLeft.transform, false);
            gameObject.transform.localPosition += new Vector3(270, 0, 0);
            //basePanel
            Image panelImage = gameObject.AddComponent<Image>();
            panelImage.material = new Material(Shader.Find("UI/Default-Overlay"));
            panelImage.color = color ?? new Color(.9f, .9f, .9f, 1);
            panelImage.rectTransform.sizeDelta = new Vector2(width, height);

            GridLayoutGroup glg = gameObject.AddComponent<GridLayoutGroup>();
            glg.cellSize = new Vector2(250, 250);
            
            glg.constraint = GridLayoutGroup.Constraint.FixedRowCount;
            glg.constraintCount = 1;
            glg.padding = new RectOffset(10,10,10,10);
            glg.spacing = new Vector2(10,0);


            AddButton.button.onClick.AddListener(AddDecalPanel);
            RemoveButton.button.onClick.AddListener(RemoveDecalPanel);

        }

        public void CreatePanels()
        {
            //Torso panel
            TorsoPanel = new ImagePanel(_mVR, null, "torso", MaterialSlot, isNormalMap, linear);
            TorsoPanel.gameObject.transform.SetParent(gameObject.transform, false);
            

            //Genital panel
            GenitalPanel = new ImagePanel(_mVR, null, "genitals", MaterialSlot, isNormalMap, linear);
            GenitalPanel.gameObject.transform.SetParent(gameObject.transform, false);
            // GenitalPanel.image.transform.localPosition = TorsoPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            //Face panel
            FacelPanel = new ImagePanel(_mVR, null, "face", MaterialSlot, isNormalMap, linear);
            FacelPanel.gameObject.transform.SetParent(gameObject.transform, false);
            //FacelPanel.image.transform.localPosition = GenitalPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            //Limbs panel
            LimbsPanel = new ImagePanel(_mVR, null, "limbs", MaterialSlot, isNormalMap, linear);
            LimbsPanel.gameObject.transform.SetParent(gameObject.transform, false);
            //LimbsPanel.image.transform.localPosition = FacelPanel.gameObject.transform.localPosition + new Vector3(270, 0, 0);

            ActivePanel = TorsoPanel;
            Selections = new Dictionary<string, ImagePanel>() { { "torso", TorsoPanel }, { "genitals", GenitalPanel }, { "face", FacelPanel }, { "limbs", LimbsPanel } };
            foreach (var sel in Selections)
            {
                sel.Value.Button.onClick.AddListener(() =>
                {
                    foreach (var b in Selections)
                    {
                        if (b.Key == sel.Key)
                        {
                            b.Value.Button.image.color = Color.white;
                            b.Value.ActivateWindow(true);
                            ActivePanel = b.Value;
                        }
                        else
                        {
                            b.Value.Button.image.color = Color.grey;
                            b.Value.ActivateWindow(false);
                        }
                    }
                });
            }
            ActivateWindow(false);
        }

        public void ActivateWindow(bool value)
        {
            spacerLeft.gameObject.SetActive(value);
            spacerRight.gameObject.SetActive(value);
            AddButton.gameObject.SetActive(value);
            RemoveButton.gameObject.SetActive(value);
            ActivePanel.ActivateWindow(value);
        }

        public void UpdateSelectionPanels()
        {
            _mVR.UpdateSelectionPanels(this);
        }
        public void ResetTextures()
        {
            TorsoPanel.ResetTextures();
            GenitalPanel.ResetTextures();
            FacelPanel.ResetTextures();
            LimbsPanel.ResetTextures();
        }

        public JSONClass Save()
        {
            JSONClass jc = new JSONClass();
            foreach (var selPanel in Selections)
            {
                JSONArray Panel = new JSONArray();

                if (selPanel.Value.DecalPanels.Count > 0)
                {
                    foreach (DecalPanel d in selPanel.Value.DecalPanels)
                    {
                        JSONClass tempDecal = d.Save();
                        Panel.Add("", tempDecal);
                    }
                   
                }
                jc[selPanel.Key] = Panel;
            }
            return jc;
        }
        public void Bake()
        {
            //SuperController.LogError("baking selectionPanel " + customShader.name);
            TorsoPanel.Bake();
            GenitalPanel.Bake();
            FacelPanel.Bake();
            LimbsPanel.Bake();
        }
        public void AddDecalPanel()
        {
            DecalPanel d = new DecalPanel(ActivePanel.spacerLeft.transform, 1050, 300, UnityEngine.Random.ColorHSV(0f, 1f, 1f, 1f, 0.5f, 1f), _mVR);
            d.gameObject.transform.localPosition += new Vector3(270, 0, 0);
            ActivePanel.IncreaseDecalPanel(d);
            d.LastDir = SetLastDir;

            d.ImagePanel.Button.onClick.AddListener(() =>
            {
                if (_lastDir == null)
                {
                    _lastDir = "Custom/Atom/Person/Textures";
                }
               SuperController.singleton.GetMediaPathDialog(new FileBrowserCallback(d.BrowserCallBack), "", _lastDir);
            });
        }
        
        public DecalPanel AddDecalPanel(string panelName)
        {
            ImagePanel active = null;
            switch (panelName)
            {
                case "torso":
                    active = TorsoPanel;
                    break;
                case "genitals":
                    active = GenitalPanel;
                    break;
                case "face":
                    active = FacelPanel;
                    break;
                case "limbs":
                    active = LimbsPanel;
                    break;   
            }
           
            DecalPanel d = new DecalPanel(active.spacerLeft.transform, 1050, 300, UnityEngine.Random.ColorHSV(0f, 1f, 1f, 1f, 0.5f, 1f), _mVR);
            d.gameObject.transform.localPosition += new Vector3(270, 0, 0);
            active.IncreaseDecalPanel(d);
            d.LastDir = SetLastDir;

            d.ImagePanel.Button.onClick.AddListener(() =>
            {
                SuperController.singleton.GetMediaPathDialog(new FileBrowserCallback(d.BrowserCallBack), "", "Custom/Atom/Person/Textures");
            });

            return d;
        }

        public void RemoveDecalPanel()
        {
            ActivePanel.RemoveDecalPanel();   
        }

        void SetLastDir(string s)
        {
            _lastDir = s;
        }

    }


    public class ImagePanel 
    {
        public GameObject gameObject = new GameObject("ImagePanel");

        public Texture Original { get; set; }
        public Image image { get; set; }
        public bool ImageLoading { get; set; }
        public string Path { get; set; }
        public Button Button { get; private set; }

        public string TextureSlot { get; set; }
        public string MaterialSlot { get; set; }
        //tell VAM image system how the texture should be loaded
        public bool isNormalMap { get; private set; }
        public bool linear { get; private set; }

        public List<DecalPanel> DecalPanels { get; set; } = new List<DecalPanel>();

        public UIDynamic spacerLeft;
        public UIDynamic spacerRight;

        private Decal_Maker _DM;
        public ImagePanel(Decal_Maker DM = null, Color? color = null, string textureSlot = null, string materialSlot = null, bool setNormalMap = false, bool setLinear = false )
        {
            TextureSlot = textureSlot;
            MaterialSlot = materialSlot;
            isNormalMap = setNormalMap;
            linear = setLinear;

            _DM = DM;
         
            image = gameObject.AddComponent<Image>();
            image.material = new Material(Shader.Find("UI/Default-Overlay"));
            image.color = color ?? new Color(1, 1, 1, 1f);

            RectTransform imageRect = image.rectTransform;
            imageRect.sizeDelta = new Vector2(300, 300);

            Button = gameObject.AddComponent<Button>();

            if (DM)
            {
                GameObject textObject = new GameObject("Text");
                textObject.AddComponent<RectTransform>();
                textObject.transform.SetParent(gameObject.transform, false);
                textObject.transform.localPosition = new Vector3(0, -35, 0);

                Text text = textObject.AddComponent<Text>();
                text.text = textureSlot.ToUpper() ?? "";
                text.alignment = TextAnchor.LowerCenter;
                text.color = new Color(0.196078435f, 0.196078435f, 0.196078435f, 1f);
                text.font = Resources.GetBuiltinResource<Font>("Arial.ttf");
                text.fontStyle = FontStyle.Bold;
                text.fontSize = 30;
                RectTransform component = textObject.GetComponent<RectTransform>();
                component.anchorMin = Vector2.zero;
                component.anchorMax = Vector2.one;
                component.sizeDelta = Vector2.zero;

                spacerLeft = DM.CreateSpacer(false);
                spacerLeft.height = 300;

                spacerRight = DM.CreateSpacer(true);
                spacerRight.height = 300;

                //Image i = spacerLeft.gameObject.AddComponent<Image>();
                //i.material = new Material(Shader.Find("UI/Default-Overlay"));
                //i.material.color = Color.red;

                GridLayoutGroup glg = spacerLeft.gameObject.AddComponent<GridLayoutGroup>();
                glg.cellSize = new Vector2(1050, spacerLeft.height);

                glg.constraint = GridLayoutGroup.Constraint.FixedColumnCount;
                glg.constraintCount = 1;
                //glg.padding = new RectOffset(10, 10, 10, 10);
                glg.spacing = new Vector2(0,0);

                RemoveDecalPanel();
                ActivateWindow(false);
            }
        }

        public void SetAlpha(float value)
        {
            Color newAlpha = image.color;
            newAlpha.a = value;
            image.color = newAlpha;
        }

        public void SetColor(Color color)
        {
            Color newColor = image.color;
            newColor.r = color.r;
            newColor.g = color.g;
            newColor.b = color.b;
            image.color = newColor;
        }

        public void IncreaseDecalPanel(DecalPanel d)
        { 
            DecalPanels.Add(d);
            ResizeSpacers(300);
        }

        public void RemoveDecalPanel()
        {
            if (DecalPanels.Any())
            {
                GameObject.Destroy(DecalPanels.Last().gameObject);
                DecalPanels.RemoveAt(DecalPanels.Count - 1);
            }
            ResizeSpacers(-300);
        }

        public void ResetTextures()
        {

            _DM.ResetTextures(this);
        }


        public void Bake()
        {
            if (DecalPanels.Count > 0)
            {
                _DM.Bake(DecalPanels, TextureSlot, MaterialSlot);
            }
        }

        private void ResizeSpacers(float value)
        {
            spacerLeft.height += value;
            spacerRight.height += value;
        }

        public void ActivateWindow(bool value)
        {
            spacerLeft.gameObject.SetActive(value);
            spacerRight.gameObject.SetActive(value);  
        }
        public void OnImageLoaded(ImageLoaderThreaded.QueuedImage qi)
        {
            if (qi != null)
            {
                //create new material instance so we can change texture seperatly
                image.material = new Material(Shader.Find("UI/Default-Overlay"));
                image.material.mainTexture = qi.tex;
                Path = qi.imgPath;
                
                ImageLoading = false;
            }
        }

        //Old VAM version does not give access to load texture so use VAM image loading routine instead
        //Newer can use FileManagementSecure.FileManagerSecure or loadTexture
        //Wil call OnImageLoad once finished   
        public void LoadPNG(string filePath, ImagePanel callBack)
        {
            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = filePath;
            queuedImage.forceReload = false;
            queuedImage.createMipMaps = false;
            queuedImage.isNormalMap = isNormalMap;
            queuedImage.isThumbnail = false;
            queuedImage.linear = linear;
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.compress = false;//!queuedImage.isNormalMap;

            queuedImage.callback = new ImageLoaderThreaded.ImageLoaderCallback(callBack.OnImageLoaded); // = new ImageLoaderThreaded.ImageLoaderCallback(OnImageLoaded);

            //start image load
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
            ImageLoading = true;
        }

    }

    public class DecalTypes
    {
        public string Type { get; set; }
        public string Path { get; set; }
        public string Name { get; set; }
        public string Creator { get; set; }
        public float Alpha { get; set; }
        public string TextureSlot { get; set; }
   
    }
    public static class TempManager
    {
        //truely temp
        static List<UnityEngine.Object> tempObjects;
        //need to keep until character or skin change
        static List<UnityEngine.Object> sessionObjects;

        static TempManager()
        {
            tempObjects = new List<UnityEngine.Object>();
            sessionObjects = new List<UnityEngine.Object>();
        }
        //Create Temp Material
        public static Material TempMaterial(Shader shader)
        {
            Material temp = new Material(shader);
            tempObjects.Add(temp);

            return temp;
        }

        //Create Temp Texture
        public static Texture2D TempTexture2D(int width, int height, TextureFormat format, bool mip = false, bool linear = false)
        {
            Texture2D temp = new Texture2D(width, height, format, mip, linear);
            tempObjects.Add(temp);
            //SuperController.LogError("Stored temp objects contains " + tempObjects.Count);
            return temp;
        }
        //Create Copy of existing texture
        public static Texture2D CopyTexture(Texture2D src ,bool linear = false)
        {
            Texture2D cpy = TempTexture2D(src.width, src.height, src.format, false, linear);
            Graphics.CopyTexture(src, cpy);
           
            return cpy;
        }
        
        //copy temp into session
        public static void MakeSessionObj(UnityEngine.Object obj)
        {
            if (obj == null)
                return;

            tempObjects.Remove(obj);
            sessionObjects.Add(obj);
        }

        //remove temp texture
        public static void DestroyTexture(UnityEngine.Object obj)
        {
            if (obj == null)
                return;

            tempObjects.Remove(obj);
            UnityEngine.Object.Destroy(obj);
        }

        public static void DestoryTempObjs()
        {
            foreach (var obj in tempObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            
            tempObjects = new List<UnityEngine.Object>();
            //SuperController.LogError("Stored temp objects Destoyed");
        }

        public static void DestorySessionObjects()
        {
            foreach (var obj in sessionObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            sessionObjects = new List<UnityEngine.Object>();
            //SuperController.LogError("Stored temp objects Destoyed");
        }

        public static void DestoryAll()
        {
            foreach (var obj in tempObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            foreach (var obj in sessionObjects)
            {
                UnityEngine.Object.Destroy(obj);
            }
            tempObjects = new List<UnityEngine.Object>();
            sessionObjects = new List<UnityEngine.Object>();
            //SuperController.LogError("All temp objects Destoyed");
        }
       
    }
  
}
